from ._version import version as __version__  # noqa: F401

# make sure to import mpi4py first
# this calls MPI_Init and registers mpi4py's atexit handler
from mpi4py import MPI

# this registers our custom XLA functions
import mpi4jax.cython

from .collective_ops.allreduce import Allreduce
from .collective_ops.recv import Recv
from .collective_ops.send import Send
from .collective_ops.sendrecv import Sendrecv
from .collective_ops.bcast import Bcast
from .flush import flush
from .warn import disable_omnistaging_warning

# at exit, we wait for all pending operations to finish
# this prevents deadlocks (see mpi4jax#22)
import atexit

atexit.register(flush)


__all__ = [
    "Allreduce",
    "Send",
    "Recv",
    "Sendrecv",
    "Bcast",
    "disable_omnistaging_warning",
    "flush",
]

del atexit, MPI, mpi4jax.cython



#!/usr/bin/env python

"""
Run with

$ mpirun -n <nproc> python -m pytest .
"""

# don't hog memory if running on GPU
import os

os.environ["XLA_PYTHON_CLIENT_PREALLOCATE"] = "false"

import jax  # noqa: E402
import jax.config  # noqa: E402
import jax.numpy as jnp  # noqa: E402

jax.config.enable_omnistaging()

from mpi4py import MPI  # noqa: E402
from mpi4jax import Allreduce  # noqa: E402

print(MPI.get_vendor())

rank = MPI.COMM_WORLD.Get_rank()
size = MPI.COMM_WORLD.Get_size()

print("MPI rank = ", rank)
print("MPI size = ", size)


seed = 123
m, n = 5, 60

k1, k2, k3, k4, k5 = jax.random.split(jax.random.PRNGKey(seed), 5)
A = jax.random.uniform(k1, (m, n))
x = jax.random.uniform(k2, (n,))
y = jax.random.uniform(k3, (m,))
v = jax.random.uniform(k4, (n,))
vprime = jax.random.uniform(k5, (m,))


assert n % size == 0
n_local = n // size
start_local = n_local * rank
end_local = start_local + n_local

A_local = A[:, start_local:end_local]
x_local = x[start_local:end_local]
v_local = v[start_local:end_local]

test_token = jax.lax.create_token(123)


# expected results:
Ax = A @ x
ATy_local = (A.T @ y)[start_local:end_local]

Av = A @ v
ATvprime_local = (A.T @ vprime)[start_local:end_local]


def allreduce_sum(x):
    res, token = Allreduce(x, op=MPI.SUM, comm=MPI.COMM_WORLD, token=test_token)
    return res


# this is just an identity
def allreduce_sumT(y):
    res, token = Allreduce(
        y, op=MPI.SUM, comm=MPI.COMM_WORLD, token=test_token, _transpose=True
    )
    return res


# x is distributed across ranks
# cols of A are distributed across ranks
# result is same on all ranks
def matvec_mpi(A_local, x_local):
    return allreduce_sum(A_local @ x_local)


# y is same on all ranks
# rows of A.T are distributed across ranks
# result is distributed across ranks
def matvec_transpose_mpi(A_local, y_global):
    return A_local.T @ allreduce_sumT(y_global)


def mv(x_local):
    return matvec_mpi(A_local, x_local)


def mvT(y):
    return matvec_transpose_mpi(A_local, y)


def transpose(f, x):
    def fT(y):
        return jax.linear_transpose(f, x)(y)[0]

    return fT


# tests


def test_matvec():
    res = mv(x_local)
    assert jnp.allclose(Ax, res)


def test_matvecT():
    res = mvT(y)
    assert jnp.allclose(ATy_local, res)


def test_matvec_transpose():
    lt = transpose(mv, x_local)
    res = lt(y)
    assert jnp.allclose(ATy_local, res)


def test_matvecT_transpose():
    ltT = transpose(mvT, y)
    res = ltT(x_local)
    assert jnp.allclose(Ax, res)


def test_matvec_transpose2():
    lt = transpose(mv, x_local)
    ltlt = transpose(lt, y)
    res = ltlt(x_local)
    assert jnp.allclose(Ax, res)


def test_matvecT_transpose2():
    ltT = transpose(mvT, y)
    ltltT = transpose(ltT, x_local)
    res = ltltT(y)
    assert jnp.allclose(ATy_local, res)


def test_matvec_transpose3():
    lt = transpose(mv, x_local)
    ltlt = transpose(lt, y)
    ltltlt = transpose(ltlt, x_local)
    res = ltltlt(y)
    assert jnp.allclose(ATy_local, res)


def test_matvecT_transpose3():
    ltT = transpose(mvT, y)
    ltltT = transpose(ltT, x_local)
    ltltltT = transpose(ltltT, y)
    res = ltltltT(x_local)
    assert jnp.allclose(Ax, res)


def test_matvec_jvp():
    res, jvp = jax.jvp(mv, (x_local,), (v_local,))
    assert jnp.allclose(Ax, res)
    assert jnp.allclose(Av, jvp)


def test_matvecT_jvp():
    res, jvp = jax.jvp(mvT, (y,), (vprime,))
    assert jnp.allclose(ATy_local, res)
    assert jnp.allclose(ATvprime_local, jvp)


def test_matvec_vjp():
    res, vjp_fun = jax.vjp(mv, x_local)
    (vjp,) = vjp_fun(vprime)
    assert jnp.allclose(Ax, res)
    assert jnp.allclose(ATvprime_local, vjp)


def test_matvecT_vjp():
    res, vjp_fun = jax.vjp(mvT, y)
    (jvp,) = vjp_fun(v_local)
    assert jnp.allclose(ATy_local, res)
    assert jnp.allclose(Av, jvp)



import jax


def flush():
    """Wait for all pending XLA operations"""
    # as suggested in jax#4335
    for device in jax.devices("cpu"):
        noop = jax.device_put(0, device=device) + 0
        noop.block_until_ready()



#!/usr/bin/env python

"""
Run with

$ mpirun -n <nproc> python -m pytest .
"""

import pytest

# don't hog memory if running on GPU
import os

os.environ["XLA_PYTHON_CLIENT_PREALLOCATE"] = "false"

import jax  # noqa: E402
import jax.config  # noqa: E402
import jax.numpy as jnp  # noqa: E402

jax.config.enable_omnistaging()

from mpi4py import MPI  # noqa: E402

print(MPI.get_vendor())

rank = MPI.COMM_WORLD.Get_rank()
size = MPI.COMM_WORLD.Get_size()

print("MPI rank = ", rank)
print("MPI size = ", size)


def test_allreduce():
    from mpi4jax import Allreduce

    arr = jnp.ones((3, 2))
    _arr = arr.copy()

    res, token = Allreduce(arr, op=MPI.SUM)
    assert jnp.array_equal(res, arr * size)
    assert jnp.array_equal(_arr, arr)


def test_allreduce_jit():
    from mpi4jax import Allreduce

    arr = jnp.ones((3, 2))
    _arr = arr.copy()

    res = jax.jit(lambda x: Allreduce(x, op=MPI.SUM)[0])(arr)
    assert jnp.array_equal(res, arr * size)
    assert jnp.array_equal(_arr, arr)


def test_allreduce_scalar():
    from mpi4jax import Allreduce

    arr = 1
    _arr = 1

    res, token = Allreduce(arr, op=MPI.SUM)
    assert jnp.array_equal(res, arr * size)
    assert jnp.array_equal(_arr, arr)


def test_allreduce_scalar_jit():
    from mpi4jax import Allreduce

    arr = 1
    _arr = 1

    res = jax.jit(lambda x: Allreduce(x, op=MPI.SUM)[0])(arr)
    assert jnp.array_equal(res, arr * size)
    assert jnp.array_equal(_arr, arr)


def test_allreduceT():
    from mpi4jax import Allreduce

    arr = jnp.ones((3, 2))
    _arr = arr.copy()

    res, _ = Allreduce(arr, op=MPI.SUM, _transpose=True)
    assert jnp.array_equal(_arr, res)


def test_allreduce_transpose():
    from mpi4jax import Allreduce

    token = jax.lax.create_token(123)
    arr = jnp.ones((3, 2))
    _arr = arr.copy()

    (res,) = jax.linear_transpose(
        lambda x: Allreduce(x, op=MPI.SUM, token=token)[0], arr
    )(_arr)
    assert jnp.array_equal(_arr, res)


def test_allreduce_transpose2():
    # test transposing twice
    from mpi4jax import Allreduce

    token = jax.lax.create_token(123)
    arr = jnp.ones((3, 2))
    _arr = arr.copy()
    _arr2 = arr.copy()

    def lt(y):
        return jax.linear_transpose(
            lambda x: Allreduce(x, op=MPI.SUM, token=token)[0], arr
        )(y)[0]

    (res,) = jax.linear_transpose(lt, _arr)(_arr2)
    expected, _ = Allreduce(_arr2, op=MPI.SUM)
    assert jnp.array_equal(expected, res)


def test_allreduceT_transpose():
    from mpi4jax import Allreduce

    token = jax.lax.create_token(123)
    arr = jnp.ones((3, 2))
    _arr = arr.copy()

    (res,) = jax.linear_transpose(
        lambda x: Allreduce(x, op=MPI.SUM, token=token, _transpose=True)[0], arr
    )(_arr)
    expected, _ = Allreduce(_arr, op=MPI.SUM)
    assert jnp.array_equal(expected, res)


def test_allreduceT_transpose2():
    # test transposing twice
    from mpi4jax import Allreduce

    token = jax.lax.create_token(123)
    arr = jnp.ones((3, 2))
    _arr = arr.copy()
    _arr2 = arr.copy()

    def lt(y):
        return jax.linear_transpose(
            lambda x: Allreduce(x, op=MPI.SUM, token=token, _transpose=True)[0], arr
        )(y)[0]

    (res,) = jax.linear_transpose(lt, _arr)(_arr2)
    assert jnp.array_equal(_arr2, res)


def test_allreduce_grad():
    from mpi4jax import Allreduce

    arr = jnp.ones((3, 2))
    _arr = arr.copy()

    res, grad = jax.value_and_grad(lambda x: Allreduce(x, op=MPI.SUM)[0].sum())(arr)
    assert jnp.array_equal(res, arr.sum() * size)
    assert jnp.array_equal(_arr, grad)

    res, grad = jax.jit(
        jax.value_and_grad(lambda x: Allreduce(x, op=MPI.SUM)[0].sum())
    )(arr)
    assert jnp.array_equal(res, arr.sum() * size)
    assert jnp.array_equal(_arr, grad)

    def testfun(x):
        y, token = Allreduce(x, op=MPI.SUM)
        z = x + 2 * y  # noqa: F841
        res, token2 = Allreduce(x, op=MPI.SUM, token=token)
        return res.sum()

    res, grad = jax.jit(jax.value_and_grad(testfun))(arr)
    assert jnp.array_equal(res, arr.sum() * size)
    assert jnp.array_equal(_arr, grad)


def test_allreduce_jvp():
    from mpi4jax import Allreduce

    token = jax.lax.create_token(123)
    arr = jnp.ones((3, 2))
    _arr = arr.copy()

    res, jvp = jax.jvp(
        lambda x: Allreduce(x, op=MPI.SUM, token=token)[0], (arr,), (_arr,)
    )

    expected, _ = Allreduce(arr, op=MPI.SUM)
    assert jnp.array_equal(expected, res)
    expected, _ = Allreduce(_arr, op=MPI.SUM)
    assert jnp.array_equal(expected, jvp)


def test_allreduce_vjp():
    from mpi4jax import Allreduce

    token = jax.lax.create_token(123)
    arr = jnp.ones((3, 2))
    _arr = arr.copy()

    res, vjp_fun = jax.vjp(lambda x: Allreduce(x, op=MPI.SUM, token=token)[0], arr)
    (vjp,) = vjp_fun(_arr)

    expected, _ = Allreduce(arr, op=MPI.SUM)
    assert jnp.array_equal(expected, res)
    assert jnp.array_equal(_arr, vjp)


def test_allreduceT_jvp():
    from mpi4jax import Allreduce

    token = jax.lax.create_token(123)
    arr = jnp.ones((3, 2))
    _arr = arr.copy()

    res, jvp = jax.jvp(
        lambda x: Allreduce(x, op=MPI.SUM, token=token, _transpose=True)[0],
        (arr,),
        (_arr,),
    )
    assert jnp.array_equal(arr, res)
    assert jnp.array_equal(_arr, jvp)


def test_allreduceT_vjp():
    from mpi4jax import Allreduce

    token = jax.lax.create_token(123)
    arr = jnp.ones((3, 2))
    _arr = arr.copy()

    res, vjp_fun = jax.vjp(
        lambda x: Allreduce(x, op=MPI.SUM, token=token, _transpose=True)[0], arr
    )
    (vjp,) = vjp_fun(_arr)

    assert jnp.array_equal(arr, res)
    expected, _ = Allreduce(_arr, op=MPI.SUM)
    assert jnp.array_equal(expected, vjp)


def test_bcast():
    from mpi4jax import Bcast

    arr = jnp.ones((3, 2))
    _arr = arr.copy()

    if rank != 0:
        _arr = _arr * 0

    res, token = Bcast(_arr, root=0)
    assert jnp.array_equal(res, arr)
    if rank == 0:
        assert jnp.array_equal(_arr, arr)


def test_bcast_jit():
    from mpi4jax import Bcast

    arr = jnp.ones((3, 2))
    _arr = arr.copy()

    if rank != 0:
        _arr = _arr * 0

    res = jax.jit(lambda x: Bcast(x, root=0)[0])(arr)
    assert jnp.array_equal(res, arr)
    if rank == 0:
        assert jnp.array_equal(_arr, arr)


def test_bcast_scalar():
    from mpi4jax import Bcast

    arr = 1
    _arr = 1

    if rank != 0:
        _arr = _arr * 0

    res, token = Bcast(_arr, root=0)
    assert jnp.array_equal(res, arr)
    if rank == 0:
        assert jnp.array_equal(_arr, arr)


def test_bcast_scalar_jit():
    from mpi4jax import Bcast

    arr = 1
    _arr = 1

    if rank != 0:
        _arr = _arr * 0

    res = jax.jit(lambda x: Bcast(x, root=0)[0])(_arr)
    assert jnp.array_equal(res, arr)
    if rank == 0:
        assert jnp.array_equal(_arr, arr)


@pytest.mark.skipif(size < 2, reason="need at least 2 processes to test send/recv")
def test_send_recv():
    from mpi4jax import Recv, Send

    arr = jnp.ones((3, 2)) * rank
    _arr = arr.copy()

    if rank == 0:
        for proc in range(1, size):
            res, token = Recv(arr, source=proc, tag=proc)
            assert jnp.array_equal(res, jnp.ones_like(arr) * proc)
            assert jnp.array_equal(_arr, arr)
    else:
        Send(arr, 0, tag=rank)
        assert jnp.array_equal(_arr, arr)


@pytest.mark.skipif(size < 2, reason="need at least 2 processes to test send/recv")
def test_send_recv_scalar():
    from mpi4jax import Recv, Send

    arr = 1 * rank
    _arr = 1 * rank

    if rank == 0:
        for proc in range(1, size):
            res, token = Recv(arr, source=proc, tag=proc)
            assert jnp.array_equal(res, jnp.ones_like(arr) * proc)
            assert jnp.array_equal(_arr, arr)
    else:
        Send(arr, 0, tag=rank)
        assert jnp.array_equal(_arr, arr)


@pytest.mark.skipif(size < 2, reason="need at least 2 processes to test send/recv")
def test_send_recv_scalar_jit():
    from mpi4jax import Recv, Send

    arr = 1 * rank
    _arr = 1 * rank

    @jax.jit
    def send_jit(x):
        Send(x, 0, tag=rank)
        return x

    if rank == 0:
        for proc in range(1, size):
            res = jax.jit(lambda x: Recv(x, source=proc, tag=proc)[0])(arr)
            assert jnp.array_equal(res, jnp.ones_like(arr) * proc)
            assert jnp.array_equal(_arr, arr)
    else:
        send_jit(arr)
        assert jnp.array_equal(_arr, arr)


@pytest.mark.skipif(size < 2, reason="need at least 2 processes to test send/recv")
def test_send_recv_jit():
    from mpi4jax import Recv, Send

    arr = jnp.ones((3, 2)) * rank
    _arr = arr.copy()

    @jax.jit
    def send_jit(x):
        Send(x, 0, tag=rank)
        return x

    if rank == 0:
        for proc in range(1, size):
            res = jax.jit(lambda x: Recv(x, source=proc, tag=proc)[0])(arr)
            assert jnp.array_equal(res, jnp.ones_like(arr) * proc)
            assert jnp.array_equal(_arr, arr)
    else:
        send_jit(arr)
        assert jnp.array_equal(_arr, arr)


@pytest.mark.skipif(size < 2 or rank > 1, reason="Runs only on rank 0 and 1")
def test_send_recv_deadlock():
    from mpi4jax import Recv, Send

    # this deadlocks without proper token management
    @jax.jit
    def deadlock(arr):
        if rank == 0:
            # send, then receive
            token = Send(arr, 1)
            newarr, _ = Recv(arr, 1, token=token)
        else:
            # receive, then send
            newarr, token = Recv(arr, 0)
            Send(arr, 0, token=token)
        return newarr

    arr = jnp.ones(10) * rank
    arr = deadlock(arr)
    assert jnp.array_equal(arr, jnp.ones_like(arr) * (1 - rank))


@pytest.mark.skipif(size < 2, reason="need at least 2 processes to test send/recv")
def test_send_recv_status():
    from mpi4jax import Recv, Send

    arr = jnp.ones((3, 2)) * rank
    _arr = arr.copy()

    if rank == 0:
        for proc in range(1, size):
            status = MPI.Status()
            res, token = Recv(arr, source=proc, tag=proc, status=status)
            assert jnp.array_equal(res, jnp.ones_like(arr) * proc)
            assert jnp.array_equal(_arr, arr)
            assert status.Get_source() == proc
    else:
        Send(arr, 0, tag=rank)
        assert jnp.array_equal(_arr, arr)


@pytest.mark.skipif(size < 2, reason="need at least 2 processes to test send/recv")
def test_send_recv_status_jit():
    from mpi4jax import Recv, Send

    arr = jnp.ones((3, 2)) * rank
    _arr = arr.copy()

    @jax.jit
    def send_jit(x):
        Send(x, 0, tag=rank)
        return x

    if rank == 0:
        for proc in range(1, size):
            status = MPI.Status()
            res = jax.jit(lambda x: Recv(x, source=proc, tag=proc, status=status)[0])(
                arr
            )
            assert jnp.array_equal(res, jnp.ones_like(arr) * proc)
            assert jnp.array_equal(_arr, arr)
            assert status.Get_source() == proc
    else:
        send_jit(arr)
        assert jnp.array_equal(_arr, arr)


@pytest.mark.skipif(size < 2 or rank > 1, reason="Runs only on rank 0 and 1")
def test_sendrecv():
    from mpi4jax import Sendrecv

    arr = jnp.ones((3, 2)) * rank
    _arr = arr.copy()

    other = 1 - rank

    res, token = Sendrecv(arr, arr, source=other, dest=other)

    assert jnp.array_equal(res, jnp.ones_like(arr) * other)
    assert jnp.array_equal(_arr, arr)


@pytest.mark.skipif(size < 2 or rank > 1, reason="Runs only on rank 0 and 1")
def test_sendrecv_status():
    from mpi4jax import Sendrecv

    arr = jnp.ones((3, 2)) * rank
    _arr = arr.copy()

    other = 1 - rank

    status = MPI.Status()
    res, token = Sendrecv(arr, arr, source=other, dest=other, status=status)

    assert jnp.array_equal(res, jnp.ones_like(arr) * other)
    assert jnp.array_equal(_arr, arr)
    assert status.Get_source() == other


@pytest.mark.skipif(size < 2 or rank > 1, reason="Runs only on rank 0 and 1")
def test_sendrecv_status_jit():
    from mpi4jax import Sendrecv

    arr = jnp.ones((3, 2)) * rank
    _arr = arr.copy()

    other = 1 - rank

    status = MPI.Status()
    res = jax.jit(
        lambda x, y: Sendrecv(x, y, source=other, dest=other, status=status)[0]
    )(arr, arr)

    assert jnp.array_equal(res, jnp.ones_like(arr) * other)
    assert jnp.array_equal(_arr, arr)
    assert status.Get_source() == other


@pytest.mark.skipif(size < 2 or rank > 1, reason="Runs only on rank 0 and 1")
def test_sendrecv_scalar():
    from mpi4jax import Sendrecv

    arr = 1 * rank
    _arr = arr

    other = 1 - rank

    res, token = Sendrecv(arr, arr, source=other, dest=other)

    assert jnp.array_equal(res, jnp.ones_like(arr) * other)
    assert jnp.array_equal(_arr, arr)


@pytest.mark.skipif(size < 2 or rank > 1, reason="Runs only on rank 0 and 1")
def test_sendrecv_jit():
    from mpi4jax import Sendrecv

    arr = jnp.ones((3, 2)) * rank
    _arr = arr.copy()

    other = 1 - rank

    res = jax.jit(lambda x, y: Sendrecv(x, y, source=other, dest=other)[0])(arr, arr)

    assert jnp.array_equal(res, jnp.ones_like(arr) * other)
    assert jnp.array_equal(_arr, arr)


@pytest.mark.skipif(size < 2 or rank > 1, reason="Runs only on rank 0 and 1")
def test_sendrecv_scalar_jit():
    from mpi4jax import Sendrecv

    arr = 1 * rank
    _arr = arr

    other = 1 - rank

    res = jax.jit(lambda x, y: Sendrecv(x, y, source=other, dest=other)[0])(arr, arr)

    assert jnp.array_equal(res, jnp.ones_like(arr) * other)
    assert jnp.array_equal(_arr, arr)


def run_in_subprocess(code, test_file, timeout=10):
    """Runs given code string in a subprocess"""
    import os
    import subprocess
    import sys
    from textwrap import dedent

    # this enables us to measure coverage in subprocesses
    cov_preamble = dedent(
        """
    try:
        import coverage
        coverage.process_startup()
    except ImportError:
        pass

    """
    )

    test_file.write_text(cov_preamble + code)

    proc = subprocess.run(
        [sys.executable, test_file],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        bufsize=0,
        timeout=timeout,
        universal_newlines=True,
        # passing a mostly empty env seems to be the only way to
        # force MPI to initialize again
        env=dict(
            PATH=os.environ["PATH"],
            COVERAGE_PROCESS_START="pyproject.toml",
            XLA_PYTHON_CLIENT_PREALLOCATE="false",
        ),
    )
    return proc


@pytest.mark.skipif(rank > 0, reason="Runs only on rank 0")
def test_abort_on_error(tmp_path):
    from textwrap import dedent

    # test in subprocess so we don't kill the testing process itself
    test_script = dedent(
        """
        import jax
        jax.config.enable_omnistaging()
        import jax.numpy as jnp

        from mpi4py import MPI
        from mpi4jax import Send

        comm = MPI.COMM_WORLD
        assert comm.Get_size() == 1

        # send to non-existing rank
        @jax.jit
        def send_jit(x):
            Send(x, dest=100, comm=comm)

        send_jit(jnp.ones(10))
    """
    )

    proc = run_in_subprocess(test_script, tmp_path / "abort.py")
    assert proc.returncode != 0
    print(proc.stderr)
    assert "r0 | MPI_Send returned error code" in proc.stderr


@pytest.mark.skipif(rank > 0, reason="Runs only on rank 0")
def test_deadlock_on_exit(tmp_path):
    from textwrap import dedent

    # without our custom atexit handler, this would deadlock most of the time
    test_script = dedent(
        """
        import jax
        jax.config.enable_omnistaging()
        import jax.numpy as jnp

        from mpi4py import MPI
        from mpi4jax import Sendrecv

        comm = MPI.COMM_WORLD
        assert comm.Get_size() == 1

        # sendrecv to self
        jax.jit(
            lambda x: Sendrecv(sendbuf=x, recvbuf=x, source=0, dest=0, comm=comm)[0]
        )(jnp.ones(10))
    """
    )

    proc = run_in_subprocess(test_script, tmp_path / "deadlock_on_exit.py")
    assert proc.returncode == 0, proc.stderr


def test_set_debug_logging(capsys):
    from mpi4jax import Allreduce
    from mpi4jax.cython.mpi_xla_bridge import set_logging

    arr = jnp.ones((3, 2))
    set_logging(True)
    res = jax.jit(lambda x: Allreduce(x, op=MPI.SUM)[0])(arr)
    res.block_until_ready()

    captured = capsys.readouterr()
    assert captured.out.startswith(f"r{rank} | MPI_Allreduce with token")

    set_logging(False)
    res = jax.jit(lambda x: Allreduce(x, op=MPI.SUM)[0])(arr)
    res.block_until_ready()

    captured = capsys.readouterr()
    assert not captured.out



import numpy as _np
from mpi4py import MPI as _MPI

from jax import abstract_arrays, core
from jax.core import Primitive
from jax.interpreters import xla
from jax.lax import create_token
from jax.lib import xla_client

from ..utils import (
    HashableMPIType,
    _constant_s32_scalar,
    _constant_u64_scalar,
    _ops,
    _unpack_builder,
    default_primitive_impl,
    dtype_ptr,
    to_mpi_ptr,
    unpack_hashable,
    wrap_as_hashable,
)
from ..validation import enforce_types
from ..warn import warn_missing_omnistaging

# The Jax primitive
mpi_recv_p = Primitive("recv_mpi")  # Create the primitive
mpi_recv_impl = default_primitive_impl(mpi_recv_p)


# This function applies the primitive to an AST
@enforce_types(
    source=_np.integer,
    tag=_np.integer,
    comm=(_MPI.Intracomm, HashableMPIType),
    status=(type(None), _MPI.Status, HashableMPIType),
    token=(type(None), xla.Token, core.Tracer),
)
def Recv(
    x,
    source=_MPI.ANY_SOURCE,
    tag=_MPI.ANY_TAG,
    comm=_MPI.COMM_WORLD,
    status=None,
    token=None,
):
    """
    Receives the input `x` from the target rank `source` using the communicator `comm`
    which defaults to the  world comunicator, with the `tag`.
    An optional token can be passed, which is used to force jax to execute
    MPI operations in the correct order.
    This is particularly important if you are performing different Send/Recv
    operations, which might otherwise deadlock.

    Argumemnts:
        x: Array or scalar input with the desired shape and dtype.
        source: rank of the source MPI process.
        tag: Tag of this message.
        comm: The communicator (defaults to MPI.COMM_WORLD)
        status:
        token: token to force a sequential order in the operations (default=None)

    Returns:
        res: the received array or scalar
        new_token: a new, modified token, that depends on this operation.
    """
    if token is None:
        token = create_token(x)

    comm = wrap_as_hashable(comm)

    if status is not None:
        status = wrap_as_hashable(status)

    return mpi_recv_p.bind(x, token, source=source, tag=tag, comm=comm, status=status)


#  This function compiles the operation
def mpi_recv_xla_encode_cpu(c, x, token, source, tag, comm, status):
    from ..cython.mpi_xla_bridge import MPI_STATUS_IGNORE_ADDR

    warn_missing_omnistaging()

    comm = unpack_hashable(comm)
    status = unpack_hashable(status)

    c = _unpack_builder(c)
    x_shape = c.GetShape(x)
    dtype = x_shape.element_type()
    dims = x_shape.dimensions()

    # compute total number of elements in array
    _nitems = _constant_s32_scalar(c, _np.prod(dims, dtype=int))
    _dtype_ptr = dtype_ptr(dtype)

    sh = xla_client.Shape.tuple_shape(
        [xla_client.Shape.array_shape(dtype, dims), xla_client.Shape.token_shape()]
    )

    if status is None:
        _status = MPI_STATUS_IGNORE_ADDR
    else:
        _status = to_mpi_ptr(status)

    operands = (
        _nitems,
        _constant_s32_scalar(c, source),
        _constant_s32_scalar(c, tag),
        _constant_u64_scalar(c, to_mpi_ptr(comm)),
        _constant_u64_scalar(c, _dtype_ptr),
        _constant_u64_scalar(c, _status),
        token,
    )

    out = _ops.CustomCall(
        c,
        b"mpi_recv",
        operands=operands,
        shape=sh,
        has_side_effect=True,
    )

    return out


def mpi_recv_xla_encode_gpu(c, x, token, source, tag, comm, status):
    from ..cython.mpi_xla_bridge import MPI_STATUS_IGNORE_ADDR
    from ..cython.mpi_xla_bridge_gpu import build_recv_descriptor

    warn_missing_omnistaging()

    comm = unpack_hashable(comm)
    status = unpack_hashable(status)

    c = _unpack_builder(c)
    x_shape = c.GetShape(x)
    dtype = x_shape.element_type()
    dims = x_shape.dimensions()

    # compute total number of elements in array
    _nitems = _np.prod(dims, dtype=int)
    _dtype_ptr = dtype_ptr(dtype)

    sh = xla_client.Shape.tuple_shape(
        [xla_client.Shape.array_shape(dtype, dims), xla_client.Shape.token_shape()]
    )

    if status is None:
        _status = MPI_STATUS_IGNORE_ADDR
    else:
        _status = to_mpi_ptr(status)

    descriptor = build_recv_descriptor(
        _nitems,
        source,
        tag,
        to_mpi_ptr(comm),
        _dtype_ptr,
        _status,
    )

    out = _ops.CustomCall(
        c,
        b"mpi_recv",
        operands=(token,),
        shape=sh,
        opaque=descriptor,
        has_side_effect=True,
    )

    return out


# This function evaluates only the shapes during AST construction
def mpi_recv_abstract_eval(xs, token, source, tag, comm, status):
    return (
        abstract_arrays.ShapedArray(xs.shape, xs.dtype),
        abstract_arrays.abstract_token,
    )


mpi_recv_p.multiple_results = True
mpi_recv_p.def_impl(mpi_recv_impl)
mpi_recv_p.def_abstract_eval(mpi_recv_abstract_eval)

# assign to the primitive the correct encoder
xla.backend_specific_translations["cpu"][mpi_recv_p] = mpi_recv_xla_encode_cpu
xla.backend_specific_translations["gpu"][mpi_recv_p] = mpi_recv_xla_encode_gpu



from cpython.pycapsule cimport PyCapsule_New

from libc.stdint cimport int32_t, uint64_t
from libc.stdlib cimport malloc, free

from mpi4py.libmpi cimport (
    MPI_Comm,
    MPI_Op,
    MPI_Datatype,
    MPI_Status,
    MPI_Type_size,
    MPI_Comm_rank
)

from .cuda_runtime_api cimport (
    cudaMemcpy,
    cudaMemcpyHostToDevice,
    cudaMemcpyDeviceToHost,
    cudaMemcpyDeviceToDevice,
    cudaStream_t,
    cudaError_t,
    cudaMemcpyKind,
    cudaSuccess,
    cudaGetErrorName,
    cudaGetErrorString,
)

from . cimport mpi_xla_bridge
from .mpi_xla_bridge cimport abort_on_error


cpdef unicode py_string(char* c_str):
    py_str = <bytes> c_str
    return py_str.decode('UTF-8')

cpdef unicode GetErrorName(cudaError_t ierr):
    return py_string(cudaGetErrorName(ierr))

cpdef unicode GetErrorString(cudaError_t ierr):
    return py_string(cudaGetErrorString(ierr))


cdef bint COPY_TO_HOST = False

cpdef void set_copy_to_host(bint enable):
    global COPY_TO_HOST
    COPY_TO_HOST = enable


cdef inline void* checked_malloc(size_t count) nogil except *:
    cdef void* mem = malloc(count)
    if not mem:
        with gil:
            raise MemoryError(f"Failed to allocate {count} bytes")

    return mem


cdef inline cudaError_t checked_cuda_memcpy(void* dst, void* src, size_t count, cudaMemcpyKind kind) nogil except *:
    cdef cudaError_t ierr
    ierr = cudaMemcpy(dst, src, count, kind)

    if ierr != cudaSuccess:
        with gil:
            err_str = GetErrorName(ierr)
            err_des = GetErrorString(ierr)

            mystr = f"cudaMemcpy failed with the following error:\n\tError {ierr} {err_str}: {err_des}"
            raise RuntimeError(mystr)

    return ierr


#
# GPU XLA targets
#

# Allreduce

cdef struct AllreduceDescriptor:
    int32_t nitems
    MPI_Op op
    MPI_Comm comm
    MPI_Datatype dtype


cpdef bytes build_allreduce_descriptor(int32_t nitems, uint64_t op_addr, uint64_t comm_addr, uint64_t dtype_addr):
    cdef AllreduceDescriptor desc = AllreduceDescriptor(
        nitems, <MPI_Op> op_addr, <MPI_Comm> comm_addr, <MPI_Datatype> dtype_addr
    )
    return bytes((<char*> &desc)[:sizeof(AllreduceDescriptor)])


cdef void mpi_allreduce(cudaStream_t* stream, void** buffers,
                        const char* opaque, size_t opaque_len) nogil except *:
    cdef int ierr, dtype_size
    cdef size_t count

    #decode inputs
    cdef void* data = buffers[0]
    cdef void* token = buffers[1]
    cdef void* out_data = buffers[2]

    cdef void* in_buf = data
    cdef void* out_buf = out_data

    if opaque_len != sizeof(AllreduceDescriptor):
        with gil:
            raise RuntimeError("got wrong size of opaque argument")

    cdef AllreduceDescriptor* desc = <AllreduceDescriptor*>(opaque)
    cdef int32_t nitems = desc.nitems
    cdef MPI_Op op = desc.op
    cdef MPI_Comm comm = desc.comm
    cdef MPI_Datatype dtype = desc.dtype

    if COPY_TO_HOST:
        # copy memory to host
        ierr = MPI_Type_size(dtype, &dtype_size)
        abort_on_error(ierr, comm, u"Type_size")

        count = dtype_size * nitems
        in_buf = checked_malloc(count)
        out_buf = checked_malloc(count)
        checked_cuda_memcpy(in_buf, data, count, cudaMemcpyDeviceToHost)

    mpi_xla_bridge.mpi_allreduce(in_buf, out_buf, nitems, dtype, op, comm, token)

    if COPY_TO_HOST:
        # copy back to device
        checked_cuda_memcpy(out_data, out_buf, count, cudaMemcpyHostToDevice)
        free(in_buf)
        free(out_buf)

    buffers[3] = token


# Bcast

cdef struct BcastDescriptor:
    int32_t nitems
    int32_t root
    MPI_Comm comm
    MPI_Datatype dtype


cpdef bytes build_bcast_descriptor(int32_t nitems, int32_t root, uint64_t comm_addr, uint64_t dtype_addr):
    cdef BcastDescriptor desc = BcastDescriptor(
        nitems, root, <MPI_Comm> comm_addr, <MPI_Datatype> dtype_addr
    )
    return bytes((<char*> &desc)[:sizeof(BcastDescriptor)])


cdef void mpi_bcast(cudaStream_t* stream, void** buffers,
                        const char* opaque, size_t opaque_len) nogil except *:
    cdef int ierr, dtype_size, rank
    cdef size_t count

    #decode inputs
    cdef void* data = buffers[0]
    cdef void* token = buffers[1]
    cdef void* out_data = buffers[2]

    cdef void* buf = out_data

    if opaque_len != sizeof(BcastDescriptor):
        with gil:
            raise RuntimeError("got wrong size of opaque argument")

    cdef BcastDescriptor* desc = <BcastDescriptor*>(opaque)
    cdef int32_t nitems = desc.nitems
    cdef int32_t root = desc.root
    cdef MPI_Comm comm = desc.comm
    cdef MPI_Datatype dtype = desc.dtype

    ierr = MPI_Type_size(dtype, &dtype_size)
    abort_on_error(ierr, comm, u"Type_size")
    ierr = MPI_Comm_rank(comm, &rank)
    abort_on_error(ierr, comm, u"Comm_rank")

    if COPY_TO_HOST:
        # copy memory to host
        count = dtype_size * nitems
        buf = checked_malloc(count)
        if rank == root:
            checked_cuda_memcpy(buf, data, count, cudaMemcpyDeviceToHost)
    else:
        if rank == root:
            checked_cuda_memcpy(buf, data, count, cudaMemcpyDeviceToDevice)

    mpi_xla_bridge.mpi_bcast(buf, nitems, dtype, op, comm, token)

    if COPY_TO_HOST:
        # copy back to device
        checked_cuda_memcpy(out_data, buf, count, cudaMemcpyHostToDevice)
        free(buf)

    buffers[3] = token



# Send

cdef struct SendDescriptor:
    int32_t nitems
    int32_t dest
    int32_t tag
    MPI_Comm comm
    MPI_Datatype dtype


cpdef bytes build_send_descriptor(int32_t nitems, int32_t dest, int32_t tag, uint64_t comm_addr, uint64_t dtype_addr):
    cdef SendDescriptor desc = SendDescriptor(
        nitems, dest, tag, <MPI_Comm> comm_addr, <MPI_Datatype> dtype_addr
    )
    return bytes((<char*> &desc)[:sizeof(SendDescriptor)])


cdef void mpi_send(cudaStream_t* stream, void** buffers,
                   const char* opaque, size_t opaque_len) nogil except *:
    cdef int ierr, dtype_size
    cdef size_t count

    #decode inputs
    cdef void* data = buffers[0]
    cdef void* token = buffers[1]

    cdef void* sendbuf = data

    if opaque_len != sizeof(SendDescriptor):
        with gil:
            raise RuntimeError("got wrong size of opaque argument")

    cdef SendDescriptor* desc = <SendDescriptor*>(opaque)
    cdef int32_t nitems = desc.nitems
    cdef int32_t dest = desc.dest
    cdef int32_t tag = desc.tag
    cdef MPI_Comm comm = desc.comm
    cdef MPI_Datatype dtype = desc.dtype

    if COPY_TO_HOST:
        # copy memory to host
        ierr = MPI_Type_size(dtype, &dtype_size)
        abort_on_error(ierr, comm, u"Type_size")

        count = dtype_size * nitems
        sendbuf = checked_malloc(count)
        checked_cuda_memcpy(sendbuf, data, count, cudaMemcpyDeviceToHost)

    mpi_xla_bridge.mpi_send(sendbuf, nitems, dtype, dest, tag, comm, token)

    if COPY_TO_HOST:
        free(sendbuf)

    buffers[2] = token


# Recv

cdef struct RecvDescriptor:
    int32_t nitems
    int32_t source
    int32_t tag
    MPI_Comm comm
    MPI_Datatype dtype
    MPI_Status* status


cpdef bytes build_recv_descriptor(int32_t nitems, int32_t dest, int32_t tag, uint64_t comm_addr,
                                  uint64_t dtype_addr, uint64_t status_addr):
    cdef RecvDescriptor desc = RecvDescriptor(
        nitems, dest, tag, <MPI_Comm> comm_addr, <MPI_Datatype> dtype_addr, <MPI_Status*> status_addr
    )
    return bytes((<char*> &desc)[:sizeof(RecvDescriptor)])


cdef void mpi_recv(cudaStream_t* stream, void** buffers,
                   const char* opaque, size_t opaque_len) nogil except *:
    cdef int ierr, dtype_size
    cdef size_t count

    #decode inputs
    cdef void* token = buffers[0]
    cdef void* out_buf = buffers[1]

    cdef void* recvbuf = out_buf

    if opaque_len != sizeof(RecvDescriptor):
        with gil:
            raise RuntimeError("got wrong size of opaque argument")

    cdef RecvDescriptor* desc = <RecvDescriptor*>(opaque)
    cdef int32_t nitems = desc.nitems
    cdef int32_t source = desc.source
    cdef int32_t tag = desc.tag
    cdef MPI_Comm comm = desc.comm
    cdef MPI_Datatype dtype = desc.dtype
    cdef MPI_Status* status = desc.status

    if COPY_TO_HOST:
        # copy memory to host
        ierr = MPI_Type_size(dtype, &dtype_size)
        abort_on_error(ierr, comm, u"Type_size")

        count = dtype_size * nitems
        recvbuf = checked_malloc(count)

    mpi_xla_bridge.mpi_recv(recvbuf, nitems, dtype, source, tag, comm, status, token)

    if COPY_TO_HOST:
        checked_cuda_memcpy(out_buf, recvbuf, count, cudaMemcpyHostToDevice)
        free(recvbuf)

    buffers[2] = token


# Sendrecv

cdef struct SendrecvDescriptor:
    int32_t sendcount
    int32_t dest
    int32_t sendtag
    MPI_Datatype sendtype
    int32_t recvcount
    int32_t source
    int32_t recvtag
    MPI_Datatype recvtype
    MPI_Comm comm
    MPI_Status* status


cpdef bytes build_sendrecv_descriptor(int32_t sendcount, int32_t dest, int32_t sendtag, uint64_t sendtype_addr,
                                      int32_t recvcount, int32_t source, int32_t recvtag, uint64_t recvtype_addr,
                                      uint64_t comm_addr, uint64_t status_addr):
    cdef SendrecvDescriptor desc = SendrecvDescriptor(
        sendcount, dest, sendtag, <MPI_Datatype> sendtype_addr,
        recvcount, source, recvtag, <MPI_Datatype> recvtype_addr,
        <MPI_Comm> comm_addr, <MPI_Status*> status_addr
    )
    return bytes((<char*> &desc)[:sizeof(SendrecvDescriptor)])


cdef void mpi_sendrecv(cudaStream_t* stream, void** buffers,
                       const char* opaque, size_t opaque_len) nogil except *:
    cdef int ierr, send_dtype_size, recv_dtype_size
    cdef size_t bytes_send, bytes_recv

    #decode inputs
    cdef void* in_buf = buffers[0]
    cdef void* token = buffers[1]
    cdef void* out_buf = buffers[2]

    cdef void* sendbuf = in_buf
    cdef void* recvbuf = out_buf

    if opaque_len != sizeof(SendrecvDescriptor):
        with gil:
            raise RuntimeError("got wrong size of opaque argument")

    cdef SendrecvDescriptor* desc = <SendrecvDescriptor*>(opaque)
    cdef int32_t sendcount = desc.sendcount
    cdef int32_t dest = desc.dest
    cdef int32_t sendtag = desc.sendtag
    cdef MPI_Datatype sendtype = desc.sendtype
    cdef int32_t recvcount = desc.recvcount
    cdef int32_t source = desc.source
    cdef int32_t recvtag = desc.recvtag
    cdef MPI_Datatype recvtype = desc.recvtype
    cdef MPI_Comm comm = desc.comm
    cdef MPI_Status* status = desc.status

    if COPY_TO_HOST:
        # copy memory to host
        ierr = MPI_Type_size(sendtype, &send_dtype_size)
        abort_on_error(ierr, comm, u"Type_size")

        ierr = MPI_Type_size(recvtype, &recv_dtype_size)
        abort_on_error(ierr, comm, u"Type_size")

        bytes_send = send_dtype_size * sendcount
        bytes_recv = recv_dtype_size * recvcount
        sendbuf = checked_malloc(bytes_send)
        recvbuf = checked_malloc(bytes_recv)
        checked_cuda_memcpy(sendbuf, in_buf, bytes_send, cudaMemcpyDeviceToHost)

    mpi_xla_bridge.mpi_sendrecv(
        sendbuf, sendcount, sendtype, dest, sendtag,
        recvbuf, recvcount, recvtype, source, recvtag,
        comm, status, token
    )

    if COPY_TO_HOST:
        checked_cuda_memcpy(out_buf, recvbuf, bytes_recv, cudaMemcpyHostToDevice)
        free(recvbuf)

    buffers[3] = token


gpu_custom_call_targets = {}

cdef register_custom_call_target(fn_name, void* fn):
    cdef const char* name = "xla._CUSTOM_CALL_TARGET"
    gpu_custom_call_targets[fn_name] = PyCapsule_New(fn, name, NULL)


register_custom_call_target(b"mpi_allreduce", <void*>(mpi_allreduce))
register_custom_call_target(b"mpi_bcast", <void*>(mpi_bcast))
register_custom_call_target(b"mpi_send", <void*>(mpi_send))
register_custom_call_target(b"mpi_recv", <void*>(mpi_recv))
register_custom_call_target(b"mpi_sendrecv", <void*>(mpi_sendrecv))



import os
import warnings

from jax.lib import xla_client

from . import mpi_xla_bridge
from . import mpi_xla_bridge_cpu

try:
    from . import mpi_xla_bridge_gpu
except ImportError:
    HAS_GPU_EXT = False
else:
    HAS_GPU_EXT = True


def is_truthy(str_val):
    return str_val.lower() in ("true", "1", "on")


def is_falsy(str_val):
    return str_val.lower() in ("false", "0", "off")


enable_logging = is_truthy(os.environ.get("MPI4JAX_DEBUG", ""))
mpi_xla_bridge.set_logging(enable_logging)


if HAS_GPU_EXT:
    gpu_copy_behavior = os.environ.get("MPI4JAX_USE_CUDA_MPI", "")

    if is_truthy(gpu_copy_behavior):
        has_cuda_mpi = True
    elif is_falsy(gpu_copy_behavior):
        has_cuda_mpi = False
    else:
        has_cuda_mpi = False
        warn_msg = (
            "Not using CUDA-enabled MPI. "
            "If you are sure that your MPI library is built with CUDA support, "
            "set MPI4JAX_USE_CUDA_MPI=1. To silence this warning, "
            "set MPI4JAX_USE_CUDA_MPI=0."
        )
        warnings.warn(warn_msg)

    mpi_xla_bridge_gpu.set_copy_to_host(not has_cuda_mpi)


# register custom call targets
for name, fn in mpi_xla_bridge_cpu.cpu_custom_call_targets.items():
    xla_client.register_custom_call_target(name, fn, platform="cpu")

if HAS_GPU_EXT:
    for name, fn in mpi_xla_bridge_gpu.gpu_custom_call_targets.items():
        xla_client.register_custom_call_target(name, fn, platform="gpu")


del os, xla_client



import pytest


def test_enforce_types():
    from types import FunctionType

    from mpi4jax.validation import enforce_types

    @enforce_types(y=(int, str), z=FunctionType)
    def foo(x, y, z):
        pass

    # ok
    foo(1, 2, lambda x: x)
    foo("test", "test", lambda x: x)

    # not ok
    with pytest.raises(TypeError) as exc:
        foo(1, lambda x: x, lambda x: x)

    assert "expected: ['int', 'str'], got: <class 'function'>" in str(exc.value)

    with pytest.raises(TypeError) as exc:
        foo(1, 2, 3)

    assert "expected: function, got: <class 'int'>" in str(exc.value)


def test_enforce_types_generic():
    import numpy as np

    from mpi4jax.validation import enforce_types

    @enforce_types(x=np.integer)
    def foo(x):
        pass

    # ok
    foo(1)
    foo(np.uint64(1))
    foo(np.int(1))

    # not ok
    with pytest.raises(TypeError) as exc:
        foo(True)

    assert "expected: integer, got: <class 'bool'>" in str(exc.value)

    with pytest.raises(TypeError) as exc:
        foo(1.2)

    assert "expected: integer, got: <class 'float'>" in str(exc.value)


def test_enforce_types_invalid_args():
    from mpi4jax.validation import enforce_types

    def foo(x):
        pass

    with pytest.raises(ValueError) as exc:
        enforce_types(a=int)(foo)

    assert 'got unexpected argument "a"' in str(exc.value)


def test_enforce_types_tracer():
    import jax

    from mpi4jax.validation import enforce_types

    @enforce_types(x=int)
    def foo(x):
        pass

    # ok
    jax.jit(foo, static_argnums=(0,))(0)

    # not ok
    with pytest.raises(TypeError) as exc:
        jax.jit(foo)(0)

    assert "abstract tracer was passed" in str(exc.value)



import functools

import numpy as _np
from jax.interpreters import xla
from jax.lib import xla_client
from mpi4py import MPI as _MPI

_ops = xla_client.ops


def default_primitive_impl(primitive):
    return functools.partial(xla.apply_primitive, primitive)


def to_mpi_ptr(mpi_obj):
    """
    to_mpi_ptr(mpi_obj)

    Returns the ptr to the underlying C mpi object
    """
    try:
        addr = _MPI._handleof(mpi_obj)
    except NotImplementedError:
        # some objects like Status only work with addressof
        addr = _MPI._addressof(mpi_obj)

    return _np.uint64(addr)


def dtype_ptr(dtype):
    """
    dtype_ptr(dtype)

    Returns the pointer to the MPI dtype of the input numpy dtype
    """
    if dtype == _np.float32:
        _dtype = to_mpi_ptr(_MPI.FLOAT)
    elif dtype == _np.float64:
        _dtype = to_mpi_ptr(_MPI.DOUBLE)
    elif dtype == _np.complex64:
        _dtype = to_mpi_ptr(_MPI.COMPLEX)
    elif dtype == _np.complex128:
        _dtype = to_mpi_ptr(_MPI.DOUBLE_COMPLEX)
    elif dtype == _np.int32:
        _dtype = to_mpi_ptr(_MPI.INT32_T)
    elif dtype == _np.int64:
        _dtype = to_mpi_ptr(_MPI.INT64_T)

    return _dtype


# Helpers to make MPI objects hashable


class HashableMPIType:
    def __init__(self, obj):
        self.wrapped = obj

    def __hash__(self):
        return int(to_mpi_ptr(self.wrapped))


def wrap_as_hashable(obj):
    if isinstance(obj, HashableMPIType):
        return obj

    return HashableMPIType(obj)


def unpack_hashable(obj):
    if isinstance(obj, HashableMPIType):
        return obj.wrapped

    return obj


# Helper functions


def _constant_s32_scalar(c, x):
    return _ops.Constant(c, _np.int32(x))


def _constant_s64_scalar(c, x):
    return _ops.Constant(c, _np.int64(x))


def _constant_u32_scalar(c, x):
    return _ops.Constant(c, _np.uint32(x))


def _constant_u64_scalar(c, x):
    return _ops.Constant(c, _np.uint64(x))


def _unpack_builder(c):
    # If `c` is a ComputationBuilder object, extracts the underlying XlaBuilder.
    return getattr(c, "_builder", c)



import numpy as _np
from mpi4py import MPI as _MPI

from jax import abstract_arrays, core
from jax.core import Primitive
from jax.interpreters import xla  # , ad
from jax.lib import xla_client

from jax.lax import create_token
from ..utils import (
    HashableMPIType,
    _constant_s32_scalar,
    _constant_u64_scalar,
    _ops,
    _unpack_builder,
    default_primitive_impl,
    dtype_ptr,
    to_mpi_ptr,
    unpack_hashable,
    wrap_as_hashable,
)
from ..validation import enforce_types
from ..warn import warn_missing_omnistaging

# The Jax primitive
mpi_bcast_p = Primitive("bcast_mpi")  # Create the primitive
mpi_bcast_impl = default_primitive_impl(mpi_bcast_p)


# This function applies the primitive to an AST
@enforce_types(
    root=(int),
    comm=(_MPI.Intracomm, HashableMPIType),
    token=(type(None), xla.Token, core.Tracer),
)
def Bcast(x, root, comm=_MPI.COMM_WORLD, token=None):
    """
    bcast(x, op, comm=_MPI.COMM_WORLD, token=None)

    Performs the bcast operation `op` on the input `x` using the
    communicator `comm` which defaults to the world comunicator.
    An optional token can be passed, which is used to force jax to execute
    MPI operations in the correct order.

    Argumemnts:
        x: Array or scalar input.
        op: The reduction operation `MPI.Op` (e.g: MPI.SUM)
        comm: The communicator (defaults to MPI.COMM_WORLD)
        token: token to force a sequential order in the operations (default=None)

    Returns:
        res: result of the bcast operation
        new_token: a new, modified token, that depends on this operation.
            This result can be ignored if result forces a data dependency.
    """
    if token is None:
        token = create_token(x)

    comm = wrap_as_hashable(comm)
    return mpi_bcast_p.bind(x, token, root=root, comm=comm)


#  This function compiles the operation
def mpi_bcast_xla_encode(c, x, token, root, comm):
    warn_missing_omnistaging()

    comm = unpack_hashable(comm)

    c = _unpack_builder(c)
    x_shape = c.GetShape(x)
    dtype = x_shape.element_type()
    dims = x_shape.dimensions()

    # compute total number of elements in array
    _nitems = _constant_s32_scalar(c, _np.prod(dims, dtype=int))

    _dtype_ptr = dtype_ptr(dtype)

    sh = xla_client.Shape.tuple_shape(
        [xla_client.Shape.array_shape(dtype, dims), xla_client.Shape.token_shape()]
    )

    return _ops.CustomCall(
        c,
        b"mpi_bcast",
        operands=(
            _nitems,
            x,
            _constant_s32_scalar(c, root),
            _constant_u64_scalar(c, to_mpi_ptr(comm)),
            _constant_u64_scalar(c, _dtype_ptr),
            token,
        ),
        shape=sh,
        has_side_effect=True,
    )


# This function evaluates only the shapes during AST construction
def mpi_bcast_abstract_eval(xs, token, root, comm):
    return (
        abstract_arrays.ShapedArray(xs.shape, xs.dtype),
        abstract_arrays.abstract_token,
    )


# def mpi_bcast_value_and_jvp(in_args, tan_args, root, comm):
#    x, token = in_args
#    x_tan, token_tan = tan_args
#
#    res = Bcast(x, token=token, dest=dest, comm=comm)
#
#    if comm.rank == root:
#        jvp = (x_tan, token_tan)
#    else:
#        jvp = (None, token_tan)
#
#    return (res, jvp)


mpi_bcast_p.multiple_results = True
mpi_bcast_p.def_impl(mpi_bcast_impl)
mpi_bcast_p.def_abstract_eval(mpi_bcast_abstract_eval)

# ad.primitive_jvps[mpi_bcast_p] = mpi_bcast_value_and_jvp

# assign to the primitive the correct encoder
xla.backend_specific_translations["cpu"][mpi_bcast_p] = mpi_bcast_xla_encode



import warnings

omnistaging_warning_enabled = True


def disable_omnistaging_warning():
    """
    Disables the warning raised when jax omnistaging is disabled.

    """
    global omnistaging_warning_enabled
    omnistaging_warning_enabled = False


def warn_missing_omnistaging():
    import jax.config as jconfig

    if omnistaging_warning_enabled and not jconfig.omnistaging_enabled:
        warnings.warn(
            "Without omnistaging, calls to jitted MPI routines might deadlock. "
            "Make sure to introduce data dependencies on all MPI calls, or use "
            "jax.config.enable_omnistaging() before importing mpi4jax."
        )



import os

from setuptools import setup
from setuptools.extension import Extension

try:
    from Cython.Build import cythonize
except ImportError:
    HAS_CYTHON = False
else:
    HAS_CYTHON = True


#######
# Utils
def search_on_path(filenames):
    for p in get_path("PATH"):
        for filename in filenames:
            full = os.path.join(p, filename)
            if os.path.exists(full):
                return os.path.abspath(full)
    return None


def print_warning(*lines):
    print("**************************************************")
    for line in lines:
        print("*** WARNING: %s" % line)
    print("**************************************************")


def get_path(key):
    return os.environ.get(key, "").split(os.pathsep)


# /end Utils
############


def mpi_info(cmd):
    import mpi4py

    config = mpi4py.get_config()
    cmd_compile = " ".join([config["mpicc"], "-show"])
    out_stream = os.popen(cmd_compile)
    flags = out_stream.read().strip()
    flags = flags.replace(",", " ").split()

    if cmd == "compile":
        startwith = "-I"
    elif cmd == "libdirs":
        startwith = "-L"
    elif cmd == "libs":
        startwith = "-l"

    out = []
    for flag in flags:
        if flag.startswith(startwith):
            out.append(flag[2:])
    return out


################
# Cuda detection

# Taken from CUPY (MIT License)
def get_cuda_path():
    nvcc_path = search_on_path(("nvcc", "nvcc.exe"))
    cuda_path_default = None
    if nvcc_path is not None:
        cuda_path_default = os.path.normpath(
            os.path.join(os.path.dirname(nvcc_path), "..")
        )

    cuda_path = os.environ.get("CUDA_PATH", "")  # Nvidia default on Windows
    if len(cuda_path) == 0:
        cuda_path = os.environ.get("CUDA_ROOT", "")  # Nvidia default on Windows

    if len(cuda_path) > 0 and cuda_path != cuda_path_default:
        print_warning(
            "nvcc path != CUDA_PATH",
            "nvcc path: %s" % cuda_path_default,
            "CUDA_PATH: %s" % cuda_path,
        )

    if os.path.exists(cuda_path):
        _cuda_path = cuda_path
    elif cuda_path_default is not None:
        _cuda_path = cuda_path_default
    elif os.path.exists("/usr/local/cuda"):
        _cuda_path = "/usr/local/cuda"
    else:
        _cuda_path = None

    return _cuda_path


def cuda_info(cmd):
    cuda_path = get_cuda_path()
    if not cuda_path:
        return []

    if cmd == "compile":
        incdir = os.path.join(cuda_path, "include")
        if os.path.isdir(incdir):
            return [incdir]

    if cmd == "libdirs":
        libdir = os.path.join(cuda_path, "lib64")
        if os.path.isdir(libdir):
            return [libdir]
        else:
            libdir = os.path.join(cuda_path, "lib")
            if os.path.isdir(libdir):
                return [libdir]

    if cmd == "libs":
        return ["cudart"]

    return []


# /end Cuda detection
#####################

if HAS_CYTHON:
    CY_EXT = "pyx"
else:
    CY_EXT = "c"

activate_tracing = os.environ.get("MPI4JAX_TRACING", "").lower() in ("true", "1", "on")

if activate_tracing:
    macros = [("CYTHON_TRACE_NOGIL", "1")]
else:
    macros = None


EXTENSIONS = [
    Extension(
        name=f"mpi4jax.cython.{mod}",
        sources=[f"mpi4jax/cython/{mod}.{CY_EXT}"],
        include_dirs=mpi_info("compile"),
        library_dirs=mpi_info("libdirs"),
        libraries=mpi_info("libs"),
        define_macros=macros,
    )
    for mod in ("mpi_xla_bridge", "mpi_xla_bridge_cpu")
]

if cuda_info("compile"):
    EXTENSIONS.append(
        Extension(
            name="mpi4jax.cython.mpi_xla_bridge_gpu",
            sources=[f"mpi4jax/cython/mpi_xla_bridge_gpu.{CY_EXT}"],
            include_dirs=mpi_info("compile") + cuda_info("compile"),
            library_dirs=mpi_info("libdirs") + cuda_info("libdirs"),
            libraries=mpi_info("libs") + cuda_info("libs"),
            define_macros=macros,
        )
    )

if HAS_CYTHON:
    compiler_directives = {"linetrace": activate_tracing}
    EXTENSIONS = cythonize(
        EXTENSIONS,
        compiler_directives=compiler_directives,
        language_level=3,
    )


setup(
    name="mpi4jax",
    author="Filippo Vicentini",
    author_email="filippovicentini@gmail.com",
    long_description="""Jax-mpi provides integration among jax and MPI, so that
    code containing MPI calls can be correctly jit-compiled through jax.""",
    url="https://github.com/PhilipVinc/mpi4jax",
    license="MIT",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    packages=["mpi4jax", "mpi4jax.collective_ops", "mpi4jax.cython"],
    ext_modules=EXTENSIONS,
    setup_requires=[
        "setuptools>=18.0",
        "cython>=0.21",
        "mpi4py>=3.0.1",
        "setuptools_scm",
    ],
    python_requires=">=3.6",
    install_requires=["jax", "jaxlib>=0.1.55", "mpi4py>=3.0.1", "numpy"],
    extras_require={"dev": ["pytest", "black", "flake8==3.8.3", "pre-commit>=2"]},
)



from cpython.pycapsule cimport PyCapsule_New

from libc.stdint cimport int32_t, uint64_t
from libc.string cimport memcpy

from mpi4py.libmpi cimport (
    MPI_Comm,
    MPI_Datatype,
    MPI_Op,
    MPI_Status,
    MPI_Comm_rank,
    MPI_Type_size,
)

from .mpi_xla_bridge cimport abort_on_error

from . cimport mpi_xla_bridge


#
# CPU XLA targets
#

cdef void mpi_send_cpu(void** out_ptr, void** data_ptr) nogil:
    cdef int32_t nitems = (<int32_t*>(data_ptr[0]))[0]
    cdef void* sendbuf = data_ptr[1]
    cdef int32_t destination = (<int32_t*>(data_ptr[2]))[0]
    cdef int32_t tag = (<int32_t*>(data_ptr[3]))[0]
    cdef MPI_Comm comm = <MPI_Comm>((<uint64_t*>(data_ptr[4]))[0])
    cdef MPI_Datatype dtype = <MPI_Datatype>((<uint64_t*>(data_ptr[5]))[0])
    cdef void* token = data_ptr[6]

    mpi_xla_bridge.mpi_send(sendbuf, nitems, dtype, destination, tag, comm, token)
    out_ptr[0] = token


cdef void mpi_recv_cpu(void** out_ptr, void** data_ptr) nogil:
    cdef int32_t nitems = (<int32_t*>(data_ptr[0]))[0]
    cdef int32_t source = (<int32_t*>(data_ptr[1]))[0]
    cdef int32_t tag = (<int32_t*>(data_ptr[2]))[0]
    cdef MPI_Comm comm = <MPI_Comm>((<uint64_t*>(data_ptr[3]))[0])
    cdef MPI_Datatype dtype = <MPI_Datatype>((<uint64_t*>(data_ptr[4]))[0])
    cdef MPI_Status* status = <MPI_Status*>((<uint64_t*>(data_ptr[5]))[0])
    cdef void* token = data_ptr[6]

    cdef void* recvbuf = out_ptr[0]
    mpi_xla_bridge.mpi_recv(recvbuf, nitems, dtype, source, tag, comm, status, token)
    out_ptr[1] = token


cdef void mpi_sendrecv_cpu(void** out_ptr, void** data_ptr) nogil:
    cdef int32_t sendcount = (<int32_t*>(data_ptr[0]))[0]
    cdef void* sendbuf = data_ptr[1]
    cdef int32_t dest = (<int32_t*>(data_ptr[2]))[0]
    cdef int32_t sendtag = (<int32_t*>(data_ptr[3]))[0]
    cdef MPI_Datatype sendtype = <MPI_Datatype>((<uint64_t*>(data_ptr[4]))[0])

    cdef int32_t recvcount = (<int32_t*>(data_ptr[5]))[0]
    cdef int32_t source = (<int32_t*>(data_ptr[6]))[0]
    cdef int32_t recvtag = (<int32_t*>(data_ptr[7]))[0]
    cdef MPI_Datatype recvtype = <MPI_Datatype>((<uint64_t*>(data_ptr[8]))[0])

    cdef MPI_Comm comm = <MPI_Comm>((<uint64_t*>(data_ptr[9]))[0])
    cdef MPI_Status* status = <MPI_Status*>((<uint64_t*>(data_ptr[10]))[0])

    cdef void* token = data_ptr[11]

    cdef void* recvbuf = out_ptr[0]
    mpi_xla_bridge.mpi_sendrecv(
        sendbuf, sendcount, sendtype, dest, sendtag,
        recvbuf, recvcount, recvtype, source, recvtag,
        comm, status, token
    )
    out_ptr[1] = token


cdef void mpi_bcast_cpu(void** out_ptr, void** data_ptr) nogil:
    cdef int32_t nitems = (<int32_t*>(data_ptr[0]))[0]
    cdef void* send_buf = data_ptr[1]
    cdef int32_t root = (<int32_t*>(data_ptr[2]))[0]
    cdef MPI_Comm comm = <MPI_Comm>((<uint64_t*>(data_ptr[3]))[0])
    cdef MPI_Datatype dtype = <MPI_Datatype>((<uint64_t*>(data_ptr[4]))[0])
    cdef void* token = data_ptr[5]

    cdef void* out_buf = out_ptr[0]

    cdef int dtype_size, rank
    ierr = MPI_Type_size(dtype, &dtype_size)
    abort_on_error(ierr, comm, u"Type_size")
    ierr = MPI_Comm_rank(comm, &rank)
    abort_on_error(ierr, comm, u"Comm_rank")

    if rank == root:
        count = dtype_size * nitems
        memcpy(out_buf, send_buf, count)

    mpi_xla_bridge.mpi_bcast(out_buf, nitems, dtype, root, comm, token)
    out_ptr[1] = token


cdef void mpi_allreduce_cpu(void** out_ptr, void** data_ptr) nogil:
    cdef int32_t nitems = (<int32_t*>(data_ptr[0]))[0]
    cdef void* sendbuf = data_ptr[1]
    cdef MPI_Op op = <MPI_Op>((<uint64_t*>(data_ptr[2]))[0])
    cdef MPI_Comm comm = <MPI_Comm>((<uint64_t*>(data_ptr[3]))[0])
    cdef MPI_Datatype dtype = <MPI_Datatype>((<uint64_t*>(data_ptr[4]))[0])
    cdef void* token = data_ptr[5]

    cdef void* recvbuf = out_ptr[0]
    mpi_xla_bridge.mpi_allreduce(
        sendbuf, recvbuf, nitems, dtype, op, comm, token
    )
    out_ptr[1] = token


cpu_custom_call_targets = {}

cdef register_custom_call_target(fn_name, void* fn):
    cdef const char* name = "xla._CUSTOM_CALL_TARGET"
    cpu_custom_call_targets[fn_name] = PyCapsule_New(fn, name, NULL)

register_custom_call_target(b"mpi_send", <void*>(mpi_send_cpu))
register_custom_call_target(b"mpi_recv", <void*>(mpi_recv_cpu))
register_custom_call_target(b"mpi_sendrecv", <void*>(mpi_sendrecv_cpu))
register_custom_call_target(b"mpi_bcast", <void*>(mpi_bcast_cpu))
register_custom_call_target(b"mpi_allreduce", <void*>(mpi_allreduce_cpu))



import sys

from libc.stdint cimport int32_t, uint64_t

cimport mpi4py.libmpi as libmpi
from mpi4py.libmpi cimport (
    MPI_Comm,
    MPI_Op,
    MPI_Datatype,
    MPI_Status,
    MPI_STATUS_IGNORE,
    MPI_SUCCESS,
    MPI_Comm_rank,
    MPI_Abort,
)

# MPI_STATUS_IGNORE is not exposed to Python by mpi4py, so we
# export its memory address as Python int here.
# This can then be passed to all functions that expect
# MPI_Status* instead of a pointer to a real status object.
MPI_STATUS_IGNORE_ADDR = int(<uint64_t>MPI_STATUS_IGNORE)


#
# Logging
#

cdef bint PRINT_DEBUG = False


cpdef void set_logging(bint enable):
    global PRINT_DEBUG
    PRINT_DEBUG = enable


cdef inline void print_debug(unicode message, MPI_Comm comm) nogil:
    cdef int rank
    MPI_Comm_rank(comm, &rank)
    with gil:
        print(f"r{rank} | {message}")


#
# Error handling
#

cdef inline int abort_on_error(int ierr, MPI_Comm comm, unicode mpi_op) nogil:
    if ierr == MPI_SUCCESS:
        return 0

    cdef int rank
    MPI_Comm_rank(comm, &rank)

    with gil:
        sys.stderr.write(
            f'r{rank} | MPI_{mpi_op} returned error code {ierr} - aborting\n'
        )
        sys.stderr.flush()
        return MPI_Abort(comm, ierr)


#
# Wrapped MPI primitives
#

cdef void mpi_send(void* sendbuf, int32_t nitems, MPI_Datatype dtype,
                   int32_t destination, int32_t tag, MPI_Comm comm, void* token) nogil:
    if PRINT_DEBUG:
        with gil:
            print_debug(f"MPI_Send -> {destination} with tag {tag} and token {<uint64_t>token:x}", comm)

    # MPI Call
    cdef int ierr
    ierr = libmpi.MPI_Send(sendbuf, nitems, dtype, destination, tag, comm)
    abort_on_error(ierr, comm, u"Send")


cdef void mpi_recv(void* recvbuf, int32_t nitems, MPI_Datatype dtype, int32_t source,
                   int32_t tag, MPI_Comm comm, MPI_Status* status, void* token) nogil:
    if PRINT_DEBUG:
        with gil:
            print_debug(f"MPI_Recv <- {source} with tag {tag} and token {<uint64_t> token:x}", comm)

    # MPI Call
    cdef int ierr
    ierr = libmpi.MPI_Recv(recvbuf, nitems, dtype, source, tag, comm, status)
    abort_on_error(ierr, comm, u"Recv")


cdef void mpi_sendrecv(void* sendbuf, int32_t sendcount, MPI_Datatype sendtype, int32_t dest, int32_t sendtag,
                       void* recvbuf, int32_t recvcount, MPI_Datatype recvtype, int32_t source, int32_t recvtag,
                       MPI_Comm comm, MPI_Status* status, void* token) nogil:
    cdef int ierr

    if PRINT_DEBUG:
        with gil:
            print_debug(
                f"MPI_Sendrecv <- {source} (tag {recvtag}) / -> {dest} (tag {sendtag}) "
                f"with token {<uint64_t>token:x}", comm
            )

    # MPI Call
    ierr = libmpi.MPI_Sendrecv(
        sendbuf, sendcount, sendtype, dest, sendtag,
        recvbuf, recvcount, recvtype, source, recvtag,
        comm, status
    )
    abort_on_error(ierr, comm, u"Sendrecv")


cdef void mpi_bcast(void* sendrecvbuf, int32_t nitems, MPI_Datatype dtype,
                   int32_t root, MPI_Comm comm, void* token) nogil:
    if PRINT_DEBUG:
        with gil:
            print_debug(f"MPI_Bcast -> {root} with token {<uint64_t>token:x}", comm)

    # MPI Call
    cdef int ierr
    ierr = libmpi.MPI_Bcast(sendrecvbuf, nitems, dtype, root, comm)
    abort_on_error(ierr, comm, u"Bcast")


cdef void mpi_allreduce(void* sendbuf, void* recvbuf, int32_t nitems,
                        MPI_Datatype dtype, MPI_Op op, MPI_Comm comm, void* token) nogil:
    cdef int ierr

    if PRINT_DEBUG:
        with gil:
            print_debug(f"MPI_Allreduce with token {<uint64_t>token:x}", comm)

    # MPI Call
    ierr = libmpi.MPI_Allreduce(sendbuf, recvbuf, nitems, dtype, op, comm)
    abort_on_error(ierr, comm, u"Allreduce")



import numpy as _np
from mpi4py import MPI as _MPI

from jax import abstract_arrays, core
from jax.core import Primitive
from jax.interpreters import ad, xla
from jax.lib import xla_client

from jax.lax import create_token
from ..utils import (
    HashableMPIType,
    _constant_s32_scalar,
    _constant_u64_scalar,
    _ops,
    _unpack_builder,
    default_primitive_impl,
    dtype_ptr,
    to_mpi_ptr,
    unpack_hashable,
    wrap_as_hashable,
)
from ..validation import enforce_types
from ..warn import warn_missing_omnistaging

# The Jax primitive
mpi_allreduce_p = Primitive("allreduce_mpi")  # Create the primitive
mpi_allreduce_impl = default_primitive_impl(mpi_allreduce_p)


# This function applies the primitive to an AST
@enforce_types(
    op=(_MPI.Op, HashableMPIType),
    comm=(_MPI.Intracomm, HashableMPIType),
    token=(type(None), xla.Token, core.Tracer),
)
def Allreduce(x, op, comm=_MPI.COMM_WORLD, token=None, _transpose=False):
    """
    Performs the Allreduce operation `op` on the input `x` using the
    communicator `comm` which defaults to the world comunicator.
    An optional token can be passed, which is used to force jax to execute
    MPI operations in the correct order.

    Arguments:
        x: Array or scalar input.
        op: The reduction operation `MPI.Op` (e.g: MPI.SUM)
        comm: The communicator (defaults to MPI.COMM_WORLD)
        token: token to force a sequential order in the operations (default=None)

    Returns:
        res: result of the allreduce operation
        new_token: a new, modified token, that depends on this operation.
            This result can be ignored if result forces a data dependency.
    """

    # The extra argument _transpose is an implementation detail. It is used to
    # keep track of whever we are computing the forward or transposition of
    # Allreduce, because we are 'cheating' and not performing MPI operations
    # on the tranposed (even though, in principle, we should) in order
    # to be more efficient.

    if token is None:
        token = create_token(x)

    op = wrap_as_hashable(op)
    comm = wrap_as_hashable(comm)
    return mpi_allreduce_p.bind(x, token, op=op, comm=comm, transpose=_transpose)


# This function compiles the operation
# transpose is a boolean flag that signals whever this is the forward pass
# performing the MPI reduction, or the transposed pass, which is trivial
def mpi_allreduce_xla_encode_cpu(c, x, token, op, comm, transpose):
    warn_missing_omnistaging()

    op = unpack_hashable(op)
    comm = unpack_hashable(comm)

    c = _unpack_builder(c)
    x_shape = c.GetShape(x)
    dtype = x_shape.element_type()
    dims = x_shape.dimensions()

    # compute total number of elements in array
    _nitems = _constant_s32_scalar(c, _np.prod(dims, dtype=int))

    _dtype_ptr = dtype_ptr(dtype)

    sh = xla_client.Shape.tuple_shape(
        [xla_client.Shape.array_shape(dtype, dims), xla_client.Shape.token_shape()]
    )

    if transpose:
        if op != _MPI.SUM:
            raise NotImplementedError(
                "The linear transpose of Allreduce for {} is not defined".format(op)
            )

        return _ops.Tuple(c, [x, token])
    else:
        return _ops.CustomCall(
            c,
            b"mpi_allreduce",
            operands=(
                _nitems,
                x,
                _constant_u64_scalar(c, to_mpi_ptr(op)),
                _constant_u64_scalar(c, to_mpi_ptr(comm)),
                _constant_u64_scalar(c, _dtype_ptr),
                token,
            ),
            shape=sh,
            has_side_effect=True,
        )


def mpi_allreduce_xla_encode_gpu(c, x, token, op, comm, transpose):
    from mpi4jax.cython import HAS_GPU_EXT

    if not HAS_GPU_EXT:
        raise RuntimeError(
            "mpi4jax GPU extensions failed to build, "
            "so it cannot be used in GPU contexts"
        )

    from mpi4jax.cython.mpi_xla_bridge_gpu import build_allreduce_descriptor

    warn_missing_omnistaging()

    op = unpack_hashable(op)
    comm = unpack_hashable(comm)

    c = _unpack_builder(c)
    x_shape = c.GetShape(x)
    dtype = x_shape.element_type()
    dims = x_shape.dimensions()

    # compute total number of elements in array
    _nitems = _np.prod(dims, dtype=int)

    _dtype_ptr = dtype_ptr(dtype)

    sh = xla_client.Shape.tuple_shape(
        [xla_client.Shape.array_shape(dtype, dims), xla_client.Shape.token_shape()]
    )

    descriptor = build_allreduce_descriptor(
        _nitems,
        to_mpi_ptr(op),
        to_mpi_ptr(comm),
        _dtype_ptr,
    )

    if transpose:
        if op != _MPI.SUM:
            raise NotImplementedError(
                "The linear transpose of Allreduce for {} is not defined".format(op)
            )

        return _ops.Tuple(c, [x, token])
    else:
        return _ops.CustomCall(
            c,
            b"mpi_allreduce",
            operands=(
                x,
                token,
            ),
            shape=sh,
            opaque=descriptor,
            has_side_effect=True,
        )


# This function evaluates only the shapes during AST construction
def mpi_allreduce_abstract_eval(xs, token, op, comm, transpose):
    return (
        abstract_arrays.ShapedArray(xs.shape, xs.dtype),
        abstract_arrays.abstract_token,
    )


def mpi_allreduce_value_and_jvp(in_args, tan_args, op, comm, transpose):
    x, token = in_args
    x_tan, token_tan = tan_args

    res = mpi_allreduce_p.bind(x, token, op=op, comm=comm, transpose=transpose)

    # Identify the correct adjoint
    op_ = unpack_hashable(op)

    if op_ == _MPI.SUM:
        jvp = mpi_allreduce_p.bind(x_tan, token, op=op, comm=comm, transpose=transpose)
    else:
        raise NotImplementedError(
            "The adjoint of allreduce for {} operation is not defined".format(op_)
        )

    return (res, jvp)


def mpi_allreduce_transpose_rule(tan_args, *x_args, op, comm, transpose):
    _, token = x_args
    t, _ = tan_args

    # Identify the correct adjoint
    op_ = unpack_hashable(op)

    if op_ == _MPI.SUM:
        return mpi_allreduce_p.bind(
            t, token, op=op, comm=comm, transpose=(not transpose)
        )
    else:
        raise NotImplementedError(
            "The linear transpose of allreduce for {} is not defined".format(op_)
        )


mpi_allreduce_p.multiple_results = True
mpi_allreduce_p.def_impl(mpi_allreduce_impl)
mpi_allreduce_p.def_abstract_eval(mpi_allreduce_abstract_eval)

ad.primitive_jvps[mpi_allreduce_p] = mpi_allreduce_value_and_jvp
ad.primitive_transposes[mpi_allreduce_p] = mpi_allreduce_transpose_rule

# assign to the primitive the correct encoder
xla.backend_specific_translations["cpu"][mpi_allreduce_p] = mpi_allreduce_xla_encode_cpu
xla.backend_specific_translations["gpu"][mpi_allreduce_p] = mpi_allreduce_xla_encode_gpu



import numpy as _np
from mpi4py import MPI as _MPI

from jax import abstract_arrays, core
from jax.core import Primitive
from jax.interpreters import xla
from jax.lax import create_token
from jax.lib import xla_client

from ..utils import (
    HashableMPIType,
    _constant_s32_scalar,
    _constant_u64_scalar,
    _ops,
    _unpack_builder,
    default_primitive_impl,
    dtype_ptr,
    to_mpi_ptr,
    unpack_hashable,
    wrap_as_hashable,
)
from ..validation import enforce_types
from ..warn import warn_missing_omnistaging

# The Jax primitive
mpi_sendrecv_p = Primitive("sendrecv_mpi")  # Create the primitive
mpi_sendrecv_impl = default_primitive_impl(mpi_sendrecv_p)


# This function applies the primitive to an AST
@enforce_types(
    source=_np.integer,
    dest=_np.integer,
    sendtag=_np.integer,
    recvtag=_np.integer,
    comm=(_MPI.Intracomm, HashableMPIType),
    status=(type(None), _MPI.Status, HashableMPIType),
    token=(type(None), xla.Token, core.Tracer),
)
def Sendrecv(
    sendbuf,
    recvbuf,
    source,
    dest,
    sendtag=0,
    recvtag=_MPI.ANY_TAG,
    comm=_MPI.COMM_WORLD,
    status=None,
    token=None,
):
    if token is None:
        token = create_token(sendbuf)

    comm = wrap_as_hashable(comm)

    if status is not None:
        status = wrap_as_hashable(status)

    return mpi_sendrecv_p.bind(
        sendbuf,
        recvbuf,
        token,
        source=source,
        dest=dest,
        sendtag=sendtag,
        recvtag=recvtag,
        comm=comm,
        status=status,
    )


#  This function compiles the operation
def mpi_sendrecv_xla_encode_cpu(
    c, sendbuf, recvbuf, token, source, dest, sendtag, recvtag, comm, status
):
    from ..cython.mpi_xla_bridge import MPI_STATUS_IGNORE_ADDR

    warn_missing_omnistaging()

    comm = unpack_hashable(comm)
    status = unpack_hashable(status)

    c = _unpack_builder(c)

    recv_shape = c.GetShape(recvbuf)
    recv_dtype = recv_shape.element_type()
    recv_dims = recv_shape.dimensions()

    # compute total number of elements in array
    _recv_nitems = _constant_s32_scalar(c, _np.prod(recv_dims, dtype=int))
    _recv_dtype_ptr = dtype_ptr(recv_dtype)

    send_shape = c.GetShape(sendbuf)
    send_dtype = send_shape.element_type()
    send_dims = send_shape.dimensions()

    # compute total number of elements in array
    _send_nitems = _constant_s32_scalar(c, _np.prod(send_dims, dtype=int))
    _send_dtype_ptr = dtype_ptr(send_dtype)

    sh = xla_client.Shape.tuple_shape(
        [
            xla_client.Shape.array_shape(recv_dtype, recv_dims),
            xla_client.Shape.token_shape(),
        ]
    )

    if status is None:
        _status = MPI_STATUS_IGNORE_ADDR
    else:
        _status = to_mpi_ptr(status)

    operands = (
        _send_nitems,
        sendbuf,
        _constant_s32_scalar(c, dest),
        _constant_s32_scalar(c, sendtag),
        _constant_u64_scalar(c, _send_dtype_ptr),
        _recv_nitems,
        _constant_s32_scalar(c, source),
        _constant_s32_scalar(c, recvtag),
        _constant_u64_scalar(c, _recv_dtype_ptr),
        _constant_u64_scalar(c, to_mpi_ptr(comm)),
        _constant_u64_scalar(c, _status),
        token,
    )

    return _ops.CustomCall(
        c,
        b"mpi_sendrecv",
        operands=operands,
        shape=sh,
        has_side_effect=True,
    )


def mpi_sendrecv_xla_encode_gpu(
    c, sendbuf, recvbuf, token, source, dest, sendtag, recvtag, comm, status
):
    from ..cython.mpi_xla_bridge import MPI_STATUS_IGNORE_ADDR
    from ..cython.mpi_xla_bridge_gpu import build_sendrecv_descriptor

    warn_missing_omnistaging()

    comm = unpack_hashable(comm)
    status = unpack_hashable(status)

    c = _unpack_builder(c)

    recv_shape = c.GetShape(recvbuf)
    recv_dtype = recv_shape.element_type()
    recv_dims = recv_shape.dimensions()

    # compute total number of elements in recv array
    _recv_nitems = _np.prod(recv_dims, dtype=int)
    _recv_dtype_ptr = dtype_ptr(recv_dtype)

    send_shape = c.GetShape(sendbuf)
    send_dtype = send_shape.element_type()
    send_dims = send_shape.dimensions()

    # compute total number of elements in send array
    _send_nitems = _np.prod(send_dims, dtype=int)
    _send_dtype_ptr = dtype_ptr(send_dtype)

    sh = xla_client.Shape.tuple_shape(
        [
            xla_client.Shape.array_shape(recv_dtype, recv_dims),
            xla_client.Shape.token_shape(),
        ]
    )

    if status is None:
        _status = MPI_STATUS_IGNORE_ADDR
    else:
        _status = to_mpi_ptr(status)

    descriptor = build_sendrecv_descriptor(
        _send_nitems,
        dest,
        sendtag,
        _send_dtype_ptr,
        _recv_nitems,
        source,
        recvtag,
        _recv_dtype_ptr,
        to_mpi_ptr(comm),
        _status,
    )

    return _ops.CustomCall(
        c,
        b"mpi_sendrecv",
        operands=(
            sendbuf,
            token,
        ),
        shape=sh,
        opaque=descriptor,
        has_side_effect=True,
    )


# This function evaluates only the shapes during AST construction
def mpi_sendrecv_abstract_eval(
    sendbuf, recvbuf, token, source, dest, sendtag, recvtag, comm, status
):
    return (
        abstract_arrays.ShapedArray(recvbuf.shape, recvbuf.dtype),
        abstract_arrays.abstract_token,
    )


mpi_sendrecv_p.multiple_results = True
mpi_sendrecv_p.def_impl(mpi_sendrecv_impl)
mpi_sendrecv_p.def_abstract_eval(mpi_sendrecv_abstract_eval)

# assign to the primitive the correct encoder
xla.backend_specific_translations["cpu"][mpi_sendrecv_p] = mpi_sendrecv_xla_encode_cpu
xla.backend_specific_translations["gpu"][mpi_sendrecv_p] = mpi_sendrecv_xla_encode_gpu



from libc.stdint cimport int32_t, uint64_t
from mpi4py.libmpi cimport MPI_Comm, MPI_Datatype, MPI_Status, MPI_Op

cdef int abort_on_error(int ierr, MPI_Comm comm, unicode mpi_op) nogil

cdef void mpi_send(void* sendbuf, int32_t nitems, MPI_Datatype dtype,
                   int32_t destination, int32_t tag, MPI_Comm comm, void* token) nogil

cdef void mpi_recv(void* recvbuf, int32_t nitems, MPI_Datatype dtype, int32_t source,
                   int32_t tag, MPI_Comm comm, MPI_Status* status, void* token) nogil

cdef void mpi_sendrecv(void* sendbuf, int32_t sendcount, MPI_Datatype sendtype, int32_t dest, int32_t sendtag,
                       void* recvbuf, int32_t recvcount, MPI_Datatype recvtype, int32_t source, int32_t recvtag,
                       MPI_Comm comm, MPI_Status* status, void* token) nogil

cdef void mpi_bcast(void* sendrecvbuf, int32_t nitems, MPI_Datatype dtype,
                   int32_t root, MPI_Comm comm, void* token) nogil

cdef void mpi_allreduce(void* sendbuf, void* recvbuf, int32_t nitems,
                        MPI_Datatype dtype, MPI_Op op, MPI_Comm comm, void* token) nogil
